@@ -22,27 +22,20 @@
 #include <linux/netdevice.h>
 #include <linux/etherdevice.h>
 #include <linux/ethtool.h>
+#include <linux/netdev_features.h>
 #include <linux/skbuff.h>
 #include <linux/version.h>
 
 #include <net/dst.h>
 #include <net/xfrm.h>
 
-#include "checksum.h"
 #include "datapath.h"
 #include "vlan.h"
 #include "vport-internal_dev.h"
 #include "vport-netdev.h"
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,1,0)
-#define HAVE_NET_DEVICE_OPS
-#endif
-
 struct internal_dev {
 	struct vport *vport;
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
-	struct net_device_stats stats;
-#endif
 };
 
 static struct internal_dev *internal_dev_priv(struct net_device *netdev)
@@ -58,11 +51,7 @@ static struct rtnl_link_stats64 *internal_dev_get_stats(struct net_device *netde
 #else
 static struct net_device_stats *internal_dev_sys_stats(struct net_device *netdev)
 {
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
-	struct net_device_stats *stats = &internal_dev_priv(netdev)->stats;
-#else
 	struct net_device_stats *stats = &netdev->stats;
-#endif
 #endif
 	struct vport *vport = ovs_internal_dev_get_vport(netdev);
 	struct ovs_vport_stats vport_stats;
@@ -86,16 +75,8 @@ static struct net_device_stats *internal_dev_sys_stats(struct net_device *netdev
 /* Called with rcu_read_lock_bh. */
 static int internal_dev_xmit(struct sk_buff *skb, struct net_device *netdev)
 {
-	if (unlikely(compute_ip_summed(skb, true))) {
-		kfree_skb(skb);
-		return 0;
-	}
-
-	vlan_copy_skb_tci(skb);
-	OVS_CB(skb)->flow = NULL;
-
 	rcu_read_lock();
-	ovs_vport_receive(internal_dev_priv(netdev)->vport, skb);
+	ovs_vport_receive(internal_dev_priv(netdev)->vport, skb, NULL);
 	rcu_read_unlock();
 	return 0;
 }
@@ -115,7 +96,7 @@ static int internal_dev_stop(struct net_device *netdev)
 static void internal_dev_getinfo(struct net_device *netdev,
 				 struct ethtool_drvinfo *info)
 {
-	strcpy(info->driver, "openvswitch");
+	strlcpy(info->driver, "openvswitch", sizeof(info->driver));
 }
 
 static const struct ethtool_ops internal_dev_ethtool_ops = {
@@ -148,7 +129,6 @@ static void internal_dev_destructor(struct net_device *dev)
 	free_netdev(dev);
 }
 
-#ifdef HAVE_NET_DEVICE_OPS
 static const struct net_device_ops internal_dev_netdev_ops = {
 	.ndo_open = internal_dev_open,
 	.ndo_stop = internal_dev_stop,
@@ -161,22 +141,12 @@ static const struct net_device_ops internal_dev_netdev_ops = {
 	.ndo_get_stats = internal_dev_sys_stats,
 #endif
 };
-#endif
 
 static void do_setup(struct net_device *netdev)
 {
 	ether_setup(netdev);
 
-#ifdef HAVE_NET_DEVICE_OPS
 	netdev->netdev_ops = &internal_dev_netdev_ops;
-#else
-	netdev->get_stats = internal_dev_sys_stats;
-	netdev->hard_start_xmit = internal_dev_xmit;
-	netdev->open = internal_dev_open;
-	netdev->stop = internal_dev_stop;
-	netdev->set_mac_address = eth_mac_addr;
-	netdev->change_mtu = internal_dev_change_mtu;
-#endif
 
 	netdev->priv_flags &= ~IFF_TX_SKB_SHARING;
 	netdev->priv_flags |= IFF_LIVE_ADDR_CHANGE;
@@ -187,10 +157,8 @@ static void do_setup(struct net_device *netdev)
 	netdev->features = NETIF_F_LLTX | NETIF_F_SG | NETIF_F_FRAGLIST |
 			   NETIF_F_HIGHDMA | NETIF_F_HW_CSUM | NETIF_F_TSO;
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
 	netdev->vlan_features = netdev->features;
-	netdev->features |= NETIF_F_HW_VLAN_TX;
-#endif
+	netdev->features |= NETIF_F_HW_VLAN_CTAG_TX;
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
 	netdev->hw_features = netdev->features & ~NETIF_F_LLTX;
@@ -229,16 +197,19 @@ static struct vport *internal_dev_create(const struct vport_parms *parms)
 	if (vport->port_no == OVSP_LOCAL)
 		netdev_vport->dev->features |= NETIF_F_NETNS_LOCAL;
 
+	rtnl_lock();
 	err = register_netdevice(netdev_vport->dev);
 	if (err)
 		goto error_free_netdev;
 
 	dev_set_promiscuity(netdev_vport->dev, 1);
+	rtnl_unlock();
 	netif_start_queue(netdev_vport->dev);
 
 	return vport;
 
 error_free_netdev:
+	rtnl_unlock();
 	free_netdev(netdev_vport->dev);
 error_free_vport:
 	ovs_vport_free(vport);
@@ -251,10 +222,13 @@ static void internal_dev_destroy(struct vport *vport)
 	struct netdev_vport *netdev_vport = netdev_vport_priv(vport);
 
 	netif_stop_queue(netdev_vport->dev);
+	rtnl_lock();
 	dev_set_promiscuity(netdev_vport->dev, -1);
 
 	/* unregister_netdevice() waits for an RCU grace period. */
 	unregister_netdevice(netdev_vport->dev);
+
+	rtnl_unlock();
 }
 
 static int internal_dev_recv(struct vport *vport, struct sk_buff *skb)
@@ -263,8 +237,19 @@ static int internal_dev_recv(struct vport *vport, struct sk_buff *skb)
 	int len;
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37)
-	if (unlikely(vlan_deaccel_tag(skb)))
-		return 0;
+	if (vlan_tx_tag_present(skb)) {
+		if (unlikely(!__vlan_put_tag(skb,
+					     skb->vlan_proto,
+					     vlan_tx_tag_get(skb))))
+			return 0;
+
+		if (skb->ip_summed == CHECKSUM_COMPLETE)
+			skb->csum = csum_add(skb->csum,
+					     csum_partial(skb->data + (2 * ETH_ALEN),
+							  VLAN_HLEN, 0));
+
+		vlan_set_tci(skb, 0);
+	}
 #endif
 
 	len = skb->len;
@@ -276,34 +261,24 @@ static int internal_dev_recv(struct vport *vport, struct sk_buff *skb)
 	skb->dev = netdev;
 	skb->pkt_type = PACKET_HOST;
 	skb->protocol = eth_type_trans(skb, netdev);
-	forward_ip_summed(skb, false);
+	skb_postpull_rcsum(skb, eth_hdr(skb), ETH_HLEN);
 
 	netif_rx(skb);
 
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,29)
-	netdev->last_rx = jiffies;
-#endif
-
 	return len;
 }
 
 const struct vport_ops ovs_internal_vport_ops = {
 	.type		= OVS_VPORT_TYPE_INTERNAL,
-	.flags		= VPORT_F_REQUIRED | VPORT_F_FLOW,
 	.create		= internal_dev_create,
 	.destroy	= internal_dev_destroy,
 	.get_name	= ovs_netdev_get_name,
-	.get_ifindex	= ovs_netdev_get_ifindex,
 	.send		= internal_dev_recv,
 };
 
 int ovs_is_internal_dev(const struct net_device *netdev)
 {
-#ifdef HAVE_NET_DEVICE_OPS
 	return netdev->netdev_ops == &internal_dev_netdev_ops;
-#else
-	return netdev->open == internal_dev_open;
-#endif
 }
 
 struct vport *ovs_internal_dev_get_vport(struct net_device *netdev)
