@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008, 2009, 2010, 2011, 2012 Nicira Networks.
+ * Copyright (c) 2008, 2009, 2010, 2011, 2012 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -17,7 +17,6 @@
 #include <config.h>
 #include "mac-learning.h"
 
-#include <assert.h>
 #include <inttypes.h>
 #include <stdlib.h>
 
@@ -26,8 +25,8 @@
 #include "hash.h"
 #include "list.h"
 #include "poll-loop.h"
-#include "tag.h"
 #include "timeval.h"
+#include "unaligned.h"
 #include "util.h"
 #include "vlan-bitmap.h"
 #include "vlog.h"
@@ -49,7 +48,9 @@ static uint32_t
 mac_table_hash(const struct mac_learning *ml, const uint8_t mac[ETH_ADDR_LEN],
                uint16_t vlan)
 {
-    return hash_bytes(mac, ETH_ADDR_LEN, vlan ^ ml->secret);
+    unsigned int mac1 = get_unaligned_u32(ALIGNED_CAST(uint32_t *, mac));
+    unsigned int mac2 = get_unaligned_u16(ALIGNED_CAST(uint16_t *, mac + 4));
+    return hash_3words(mac1, mac2 | (vlan << 16), ml->secret);
 }
 
 static struct mac_entry *
@@ -58,16 +59,6 @@ mac_entry_from_lru_node(struct list *list)
     return CONTAINER_OF(list, struct mac_entry, lru_node);
 }
 
-/* Returns a tag that represents that 'mac' is on an unknown port in 'vlan'.
- * (When we learn where 'mac' is in 'vlan', this allows flows that were
- * flooded to be revalidated.) */
-static tag_type
-make_unknown_mac_tag(const struct mac_learning *ml,
-                     const uint8_t mac[ETH_ADDR_LEN], uint16_t vlan)
-{
-    return tag_create_deterministic(mac_table_hash(ml, mac, vlan));
-}
-
 static struct mac_entry *
 mac_entry_lookup(const struct mac_learning *ml,
                  const uint8_t mac[ETH_ADDR_LEN], uint16_t vlan)
@@ -88,6 +79,7 @@ mac_entry_lookup(const struct mac_learning *ml,
  * and return false. */
 static bool
 get_lru(struct mac_learning *ml, struct mac_entry **e)
+    OVS_REQ_RDLOCK(ml->rwlock)
 {
     if (!list_is_empty(&ml->lrus)) {
         *e = mac_entry_from_lru_node(ml->lrus.next);
@@ -107,7 +99,8 @@ normalize_idle_time(unsigned int idle_time)
 }
 
 /* Creates and returns a new MAC learning table with an initial MAC aging
- * timeout of 'idle_time' seconds. */
+ * timeout of 'idle_time' seconds and an initial maximum of MAC_DEFAULT_MAX
+ * entries. */
 struct mac_learning *
 mac_learning_create(unsigned int idle_time)
 {
@@ -119,14 +112,38 @@ mac_learning_create(unsigned int idle_time)
     ml->secret = random_uint32();
     ml->flood_vlans = NULL;
     ml->idle_time = normalize_idle_time(idle_time);
+    ml->max_entries = MAC_DEFAULT_MAX;
+    ml->need_revalidate = false;
+    atomic_init(&ml->ref_cnt, 1);
+    ovs_rwlock_init(&ml->rwlock);
     return ml;
 }
 
-/* Destroys MAC learning table 'ml'. */
-void
-mac_learning_destroy(struct mac_learning *ml)
+struct mac_learning *
+mac_learning_ref(const struct mac_learning *ml_)
 {
+    struct mac_learning *ml = CONST_CAST(struct mac_learning *, ml_);
     if (ml) {
+        int orig;
+        atomic_add(&ml->ref_cnt, 1, &orig);
+        ovs_assert(orig > 0);
+    }
+    return ml;
+}
+
+/* Unreferences (and possibly destroys) MAC learning table 'ml'. */
+void
+mac_learning_unref(struct mac_learning *ml)
+{
+    int orig;
+
+    if (!ml) {
+        return;
+    }
+
+    atomic_sub(&ml->ref_cnt, 1, &orig);
+    ovs_assert(orig > 0);
+    if (orig == 1) {
         struct mac_entry *e, *next;
 
         HMAP_FOR_EACH_SAFE (e, next, hmap_node, &ml->table) {
@@ -136,6 +153,7 @@ mac_learning_destroy(struct mac_learning *ml)
         hmap_destroy(&ml->table);
 
         bitmap_free(ml->flood_vlans);
+        ovs_rwlock_destroy(&ml->rwlock);
         free(ml);
     }
 }
@@ -173,6 +191,16 @@ mac_learning_set_idle_time(struct mac_learning *ml, unsigned int idle_time)
     }
 }
 
+/* Sets the maximum number of entries in 'ml' to 'max_entries', adjusting it
+ * to be within a reasonable range. */
+void
+mac_learning_set_max_entries(struct mac_learning *ml, size_t max_entries)
+{
+    ml->max_entries = (max_entries < 10 ? 10
+                       : max_entries > 1000 * 1000 ? 1000 * 1000
+                       : max_entries);
+}
+
 static bool
 is_learning_vlan(const struct mac_learning *ml, uint16_t vlan)
 {
@@ -209,7 +237,7 @@ mac_learning_insert(struct mac_learning *ml,
     if (!e) {
         uint32_t hash = mac_table_hash(ml, src_mac, vlan);
 
-        if (hmap_count(&ml->table) >= MAC_MAX) {
+        if (hmap_count(&ml->table) >= ml->max_entries) {
             get_lru(ml, &e);
             mac_learning_expire(ml, e);
         }
@@ -218,8 +246,8 @@ mac_learning_insert(struct mac_learning *ml,
         hmap_insert(&ml->table, &e->hmap_node, hash);
         memcpy(e->mac, src_mac, ETH_ADDR_LEN);
         e->vlan = vlan;
-        e->tag = 0;
         e->grat_arp_lock = TIME_MIN;
+        e->port.p = NULL;
     } else {
         list_remove(&e->lru_node);
     }
@@ -231,22 +259,19 @@ mac_learning_insert(struct mac_learning *ml,
     return e;
 }
 
-/* Changes 'e''s tag to a new, randomly selected one, and returns the tag that
- * would have been previously used for this entry's MAC and VLAN (either before
- * 'e' was inserted, if it is new, or otherwise before its port was updated.)
+/* Changes 'e''s tag to a new, randomly selected one.  Causes
+ * mac_learning_run() to flag for revalidation the tag that would have been
+ * previously used for this entry's MAC and VLAN (either before 'e' was
+ * inserted, if it is new, or otherwise before its port was updated.)
  *
  * The client should call this function after obtaining a MAC learning entry
  * from mac_learning_insert(), if the entry is either new or if its learned
  * port has changed. */
-tag_type
-mac_learning_changed(struct mac_learning *ml, struct mac_entry *e)
+void
+mac_learning_changed(struct mac_learning *ml)
 {
-    tag_type old_tag = e->tag;
-
     COVERAGE_INC(mac_learning_learned);
-
-    e->tag = tag_create_random();
-    return old_tag ? old_tag : make_unknown_mac_tag(ml, e->mac, e->vlan);
+    ml->need_revalidate = true;
 }
 
 /* Looks up MAC 'dst' for VLAN 'vlan' in 'ml' and returns the associated MAC
@@ -255,8 +280,7 @@ mac_learning_changed(struct mac_learning *ml, struct mac_entry *e)
  * '*tag'. */
 struct mac_entry *
 mac_learning_lookup(const struct mac_learning *ml,
-                    const uint8_t dst[ETH_ADDR_LEN], uint16_t vlan,
-                    tag_type *tag)
+                    const uint8_t dst[ETH_ADDR_LEN], uint16_t vlan)
 {
     if (eth_addr_is_multicast(dst)) {
         /* No tag because the treatment of multicast destinations never
@@ -269,11 +293,7 @@ mac_learning_lookup(const struct mac_learning *ml,
     } else {
         struct mac_entry *e = mac_entry_lookup(ml, dst, vlan);
 
-        assert(e == NULL || e->tag != 0);
-        if (tag) {
-            /* Tag either the learned port or the lack thereof. */
-            *tag |= e ? e->tag : make_unknown_mac_tag(ml, dst, vlan);
-        }
+        ovs_assert(e == NULL || e->port.p != NULL)
         return e;
     }
 }
@@ -287,36 +307,49 @@ mac_learning_expire(struct mac_learning *ml, struct mac_entry *e)
     free(e);
 }
 
-/* Expires all the mac-learning entries in 'ml'.  The tags in 'ml' are
- * discarded, so the client is responsible for revalidating any flows that
- * depend on 'ml', if necessary. */
+/* Expires all the mac-learning entries in 'ml'.  If not NULL, the tags in 'ml'
+ * are added to 'tags'.  Otherwise the tags in 'ml' are discarded.  The client
+ * is responsible for revalidating any flows that depend on 'ml', if
+ * necessary. */
 void
 mac_learning_flush(struct mac_learning *ml)
 {
     struct mac_entry *e;
     while (get_lru(ml, &e)){
+        ml->need_revalidate = true;
         mac_learning_expire(ml, e);
     }
     hmap_shrink(&ml->table);
 }
 
-void
-mac_learning_run(struct mac_learning *ml, struct tag_set *set)
+/* Does periodic work required by 'ml'.  Returns true if something changed that
+ * may require flow revalidation. */
+bool
+mac_learning_run(struct mac_learning *ml)
 {
+    bool need_revalidate;
     struct mac_entry *e;
-    while (get_lru(ml, &e) && time_now() >= e->expires) {
+
+    while (get_lru(ml, &e)
+           && (hmap_count(&ml->table) > ml->max_entries
+               || time_now() >= e->expires)) {
         COVERAGE_INC(mac_learning_expired);
-        if (set) {
-            tag_set_add(set, e->tag);
-        }
+        ml->need_revalidate = true;
         mac_learning_expire(ml, e);
     }
+
+    need_revalidate = ml->need_revalidate;
+    ml->need_revalidate = false;
+    return need_revalidate;
 }
 
 void
 mac_learning_wait(struct mac_learning *ml)
 {
-    if (!list_is_empty(&ml->lrus)) {
+    if (hmap_count(&ml->table) > ml->max_entries
+        || ml->need_revalidate) {
+        poll_immediate_wake();
+    } else if (!list_is_empty(&ml->lrus)) {
         struct mac_entry *e = mac_entry_from_lru_node(ml->lrus.next);
         poll_timer_wait_until(e->expires * 1000LL);
     }
