@@ -66,6 +66,8 @@ state_name(enum state state)
  *
  * See the large comment in rconn.h for more information. */
 struct rconn {
+    struct ovs_mutex mutex;
+
     enum state state;
     time_t state_entered;
 
@@ -139,21 +141,71 @@ uint32_t rconn_get_allowed_versions(const struct rconn *rconn)
     return rconn->allowed_versions;
 }
 
-static unsigned int elapsed_in_this_state(const struct rconn *);
-static unsigned int timeout(const struct rconn *);
-static bool timed_out(const struct rconn *);
-static void state_transition(struct rconn *, enum state);
-static void rconn_set_target__(struct rconn *,
-                               const char *target, const char *name);
-static int try_send(struct rconn *);
-static void reconnect(struct rconn *);
-static void report_error(struct rconn *, int error);
-static void disconnect(struct rconn *, int error);
-static void flush_queue(struct rconn *);
+static unsigned int elapsed_in_this_state(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex);
+static unsigned int timeout(const struct rconn *rc) OVS_REQUIRES(rc->mutex);
+static bool timed_out(const struct rconn *rc) OVS_REQUIRES(rc->mutex);
+static void state_transition(struct rconn *rc, enum state)
+    OVS_REQUIRES(rc->mutex);
+static void rconn_set_target__(struct rconn *rc,
+                               const char *target, const char *name)
+    OVS_REQUIRES(rc->mutex);
+static int rconn_send__(struct rconn *rc, struct ofpbuf *,
+                        struct rconn_packet_counter *)
+    OVS_REQUIRES(rc->mutex);
+static int try_send(struct rconn *rc) OVS_REQUIRES(rc->mutex);
+static void reconnect(struct rconn *rc) OVS_REQUIRES(rc->mutex);
+static void report_error(struct rconn *rc, int error) OVS_REQUIRES(rc->mutex);
+static void rconn_disconnect__(struct rconn *rc) OVS_REQUIRES(rc->mutex);
+static void disconnect(struct rconn *rc, int error) OVS_REQUIRES(rc->mutex);
+static void flush_queue(struct rconn *rc) OVS_REQUIRES(rc->mutex);
+static void close_monitor(struct rconn *rc, size_t idx, int retval)
+    OVS_REQUIRES(rc->mutex);
 static void copy_to_monitor(struct rconn *, const struct ofpbuf *);
 static bool is_connected_state(enum state);
 static bool is_admitted_msg(const struct ofpbuf *);
-static bool rconn_logging_connection_attempts__(const struct rconn *);
+static bool rconn_logging_connection_attempts__(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex);
+static int rconn_get_version__(const struct rconn *rconn)
+    OVS_REQUIRES(rconn->mutex);
+
+/* The following prototypes duplicate those in rconn.h, but there we weren't
+ * able to add the OVS_EXCLUDED annotations because the definition of struct
+ * rconn was not visible. */
+
+void rconn_set_max_backoff(struct rconn *rc, int max_backoff)
+    OVS_EXCLUDED(rc->mutex);
+void rconn_connect(struct rconn *rc, const char *target, const char *name)
+    OVS_EXCLUDED(rc->mutex);
+void rconn_connect_unreliably(struct rconn *rc,
+                              struct vconn *vconn, const char *name)
+    OVS_EXCLUDED(rc->mutex);
+void rconn_reconnect(struct rconn *rc) OVS_EXCLUDED(rc->mutex);
+void rconn_disconnect(struct rconn *rc) OVS_EXCLUDED(rc->mutex);
+void rconn_run(struct rconn *rc) OVS_EXCLUDED(rc->mutex);
+void rconn_run_wait(struct rconn *rc) OVS_EXCLUDED(rc->mutex);
+struct ofpbuf *rconn_recv(struct rconn *rc) OVS_EXCLUDED(rc->mutex);
+void rconn_recv_wait(struct rconn *rc) OVS_EXCLUDED(rc->mutex);
+int rconn_send(struct rconn *rc, struct ofpbuf *b,
+               struct rconn_packet_counter *counter)
+    OVS_EXCLUDED(rc->mutex);
+int rconn_send_with_limit(struct rconn *rc, struct ofpbuf *b,
+                          struct rconn_packet_counter *counter,
+                          int queue_limit)
+    OVS_EXCLUDED(rc->mutex);
+void rconn_add_monitor(struct rconn *rc, struct vconn *vconn)
+    OVS_EXCLUDED(rc->mutex);
+void rconn_set_name(struct rconn *rc, const char *new_name)
+    OVS_EXCLUDED(rc->mutex);
+bool rconn_is_admitted(const struct rconn *rconn) OVS_EXCLUDED(rconn->mutex);
+int rconn_failure_duration(const struct rconn *rconn)
+    OVS_EXCLUDED(rconn->mutex);
+ovs_be16 rconn_get_local_port(const struct rconn *rconn)
+    OVS_EXCLUDED(rconn->mutex);
+int rconn_get_version(const struct rconn *rconn) OVS_EXCLUDED(rconn->mutex);
+unsigned int rconn_count_txqlen(const struct rconn *rc)
+    OVS_EXCLUDED(rc->mutex);
+
 
 /* Creates and returns a new rconn.
  *
@@ -183,6 +235,8 @@ rconn_create(int probe_interval, int max_backoff, uint8_t dscp,
 {
     struct rconn *rc = xzalloc(sizeof *rc);
 
+    ovs_mutex_init(&rc->mutex);
+
     rc->state = S_VOID;
     rc->state_entered = time_now();
 
@@ -224,7 +278,9 @@ rconn_create(int probe_interval, int max_backoff, uint8_t dscp,
 
 void
 rconn_set_max_backoff(struct rconn *rc, int max_backoff)
+    OVS_EXCLUDED(rc->mutex)
 {
+    ovs_mutex_lock(&rc->mutex);
     rc->max_backoff = MAX(1, max_backoff);
     if (rc->state == S_BACKOFF && rc->backoff > max_backoff) {
         rc->backoff = max_backoff;
@@ -232,6 +288,7 @@ rconn_set_max_backoff(struct rconn *rc, int max_backoff)
             rc->backoff_deadline = time_now() + max_backoff;
         }
     }
+    ovs_mutex_unlock(&rc->mutex);
 }
 
 int
@@ -273,11 +330,14 @@ rconn_get_probe_interval(const struct rconn *rc)
  * but it need not be acceptable to vconn_open(). */
 void
 rconn_connect(struct rconn *rc, const char *target, const char *name)
+    OVS_EXCLUDED(rc->mutex)
 {
-    rconn_disconnect(rc);
+    ovs_mutex_lock(&rc->mutex);
+    rconn_disconnect__(rc);
     rconn_set_target__(rc, target, name);
     rc->reliable = true;
     reconnect(rc);
+    ovs_mutex_unlock(&rc->mutex);
 }
 
 /* Drops any existing connection on 'rc', then configures 'rc' to use
@@ -291,28 +351,36 @@ rconn_connect(struct rconn *rc, const char *target, const char *name)
 void
 rconn_connect_unreliably(struct rconn *rc,
                          struct vconn *vconn, const char *name)
+    OVS_EXCLUDED(rc->mutex)
 {
     ovs_assert(vconn != NULL);
-    rconn_disconnect(rc);
+
+    ovs_mutex_lock(&rc->mutex);
+    rconn_disconnect__(rc);
     rconn_set_target__(rc, vconn_get_name(vconn), name);
     rc->reliable = false;
     rc->vconn = vconn;
     rc->last_connected = time_now();
     state_transition(rc, S_ACTIVE);
+    ovs_mutex_unlock(&rc->mutex);
 }
 
 /* If 'rc' is connected, forces it to drop the connection and reconnect. */
 void
 rconn_reconnect(struct rconn *rc)
+    OVS_EXCLUDED(rc->mutex)
 {
+    ovs_mutex_lock(&rc->mutex);
     if (rc->state & (S_ACTIVE | S_IDLE)) {
         VLOG_INFO("%s: disconnecting", rc->name);
         disconnect(rc, 0);
     }
+    ovs_mutex_unlock(&rc->mutex);
 }
 
-void
-rconn_disconnect(struct rconn *rc)
+static void
+rconn_disconnect__(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     if (rc->state != S_VOID) {
         if (rc->vconn) {
@@ -329,6 +397,15 @@ rconn_disconnect(struct rconn *rc)
     }
 }
 
+void
+rconn_disconnect(struct rconn *rc)
+    OVS_EXCLUDED(rc->mutex)
+{
+    ovs_mutex_lock(&rc->mutex);
+    rconn_disconnect__(rc);
+    ovs_mutex_unlock(&rc->mutex);
+}
+
 /* Disconnects 'rc' and frees the underlying storage. */
 void
 rconn_destroy(struct rconn *rc)
@@ -336,6 +413,7 @@ rconn_destroy(struct rconn *rc)
     if (rc) {
         size_t i;
 
+        ovs_mutex_lock(&rc->mutex);
         free(rc->name);
         free(rc->target);
         vconn_close(rc->vconn);
@@ -344,24 +422,30 @@ rconn_destroy(struct rconn *rc)
         for (i = 0; i < rc->n_monitors; i++) {
             vconn_close(rc->monitors[i]);
         }
+        ovs_mutex_unlock(&rc->mutex);
+        ovs_mutex_destroy(&rc->mutex);
+
         free(rc);
     }
 }
 
 static unsigned int
 timeout_VOID(const struct rconn *rc OVS_UNUSED)
+    OVS_REQUIRES(rc->mutex)
 {
     return UINT_MAX;
 }
 
 static void
 run_VOID(struct rconn *rc OVS_UNUSED)
+    OVS_REQUIRES(rc->mutex)
 {
     /* Nothing to do. */
 }
 
 static void
 reconnect(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     int retval;
 
@@ -378,7 +462,8 @@ reconnect(struct rconn *rc)
         rc->backoff_deadline = time_now() + rc->backoff;
         state_transition(rc, S_CONNECTING);
     } else {
-        VLOG_WARN("%s: connection failed (%s)", rc->name, strerror(retval));
+        VLOG_WARN("%s: connection failed (%s)",
+                  rc->name, ovs_strerror(retval));
         rc->backoff_deadline = TIME_MAX; /* Prevent resetting backoff. */
         disconnect(rc, retval);
     }
@@ -386,12 +471,14 @@ reconnect(struct rconn *rc)
 
 static unsigned int
 timeout_BACKOFF(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     return rc->backoff;
 }
 
 static void
 run_BACKOFF(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     if (timed_out(rc)) {
         reconnect(rc);
@@ -400,12 +487,14 @@ run_BACKOFF(struct rconn *rc)
 
 static unsigned int
 timeout_CONNECTING(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     return MAX(1, rc->backoff);
 }
 
 static void
 run_CONNECTING(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     int retval = vconn_connect(rc->vconn);
     if (!retval) {
@@ -416,7 +505,7 @@ run_CONNECTING(struct rconn *rc)
     } else if (retval != EAGAIN) {
         if (rconn_logging_connection_attempts__(rc)) {
             VLOG_INFO("%s: connection failed (%s)",
-                      rc->name, strerror(retval));
+                      rc->name, ovs_strerror(retval));
         }
         disconnect(rc, retval);
     } else if (timed_out(rc)) {
@@ -430,6 +519,7 @@ run_CONNECTING(struct rconn *rc)
 
 static void
 do_tx_work(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     if (list_is_empty(&rc->txq)) {
         return;
@@ -448,6 +538,7 @@ do_tx_work(struct rconn *rc)
 
 static unsigned int
 timeout_ACTIVE(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     if (rc->probe_interval) {
         unsigned int base = MAX(rc->last_activity, rc->state_entered);
@@ -459,6 +550,7 @@ timeout_ACTIVE(const struct rconn *rc)
 
 static void
 run_ACTIVE(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     if (timed_out(rc)) {
         unsigned int base = MAX(rc->last_activity, rc->state_entered);
@@ -467,14 +559,14 @@ run_ACTIVE(struct rconn *rc)
         VLOG_DBG("%s: idle %u seconds, sending inactivity probe",
                  rc->name, (unsigned int) (time_now() - base));
 
-        version = rconn_get_version(rc);
+        version = rconn_get_version__(rc);
         ovs_assert(version >= 0 && version <= 0xff);
 
         /* Ordering is important here: rconn_send() can transition to BACKOFF,
          * and we don't want to transition back to IDLE if so, because then we
          * can end up queuing a packet with vconn == NULL and then *boom*. */
         state_transition(rc, S_IDLE);
-        rconn_send(rc, make_echo_request(version), NULL);
+        rconn_send__(rc, make_echo_request(version), NULL);
         return;
     }
 
@@ -483,12 +575,14 @@ run_ACTIVE(struct rconn *rc)
 
 static unsigned int
 timeout_IDLE(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     return rc->probe_interval;
 }
 
 static void
 run_IDLE(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     if (timed_out(rc)) {
         VLOG_ERR("%s: no response to inactivity probe after %u "
@@ -505,15 +599,30 @@ run_IDLE(struct rconn *rc)
  * connected, attempts to send packets in the send queue, if any. */
 void
 rconn_run(struct rconn *rc)
+    OVS_EXCLUDED(rc->mutex)
 {
     int old_state;
     size_t i;
 
+    ovs_mutex_lock(&rc->mutex);
     if (rc->vconn) {
         vconn_run(rc->vconn);
     }
-    for (i = 0; i < rc->n_monitors; i++) {
+    for (i = 0; i < rc->n_monitors; ) {
+        struct ofpbuf *msg;
+        int retval;
+
         vconn_run(rc->monitors[i]);
+
+        /* Drain any stray message that came in on the monitor connection. */
+        retval = vconn_recv(rc->monitors[i], &msg);
+        if (!retval) {
+            ofpbuf_delete(msg);
+        } else if (retval != EAGAIN) {
+            close_monitor(rc, i, retval);
+            continue;
+        }
+        i++;
     }
 
     do {
@@ -526,16 +635,19 @@ rconn_run(struct rconn *rc)
             NOT_REACHED();
         }
     } while (rc->state != old_state);
+    ovs_mutex_unlock(&rc->mutex);
 }
 
 /* Causes the next call to poll_block() to wake up when rconn_run() should be
  * called on 'rc'. */
 void
 rconn_run_wait(struct rconn *rc)
+    OVS_EXCLUDED(rc->mutex)
 {
     unsigned int timeo;
     size_t i;
 
+    ovs_mutex_lock(&rc->mutex);
     if (rc->vconn) {
         vconn_run_wait(rc->vconn);
         if ((rc->state & (S_ACTIVE | S_IDLE)) && !list_is_empty(&rc->txq)) {
@@ -544,6 +656,7 @@ rconn_run_wait(struct rconn *rc)
     }
     for (i = 0; i < rc->n_monitors; i++) {
         vconn_run_wait(rc->monitors[i]);
+        vconn_recv_wait(rc->monitors[i]);
     }
 
     timeo = timeout(rc);
@@ -551,6 +664,7 @@ rconn_run_wait(struct rconn *rc)
         long long int expires = sat_add(rc->state_entered, timeo);
         poll_timer_wait_until(expires * 1000);
     }
+    ovs_mutex_unlock(&rc->mutex);
 }
 
 /* Attempts to receive a packet from 'rc'.  If successful, returns the packet;
@@ -558,9 +672,12 @@ rconn_run_wait(struct rconn *rc)
  * the packet (with ofpbuf_delete()). */
 struct ofpbuf *
 rconn_recv(struct rconn *rc)
+    OVS_EXCLUDED(rc->mutex)
 {
+    struct ofpbuf *buffer = NULL;
+
+    ovs_mutex_lock(&rc->mutex);
     if (rc->state & (S_ACTIVE | S_IDLE)) {
-        struct ofpbuf *buffer;
         int error = vconn_recv(rc->vconn, &buffer);
         if (!error) {
             copy_to_monitor(rc, buffer);
@@ -574,40 +691,33 @@ rconn_recv(struct rconn *rc)
             if (rc->state == S_IDLE) {
                 state_transition(rc, S_ACTIVE);
             }
-            return buffer;
         } else if (error != EAGAIN) {
             report_error(rc, error);
             disconnect(rc, error);
         }
     }
-    return NULL;
+    ovs_mutex_unlock(&rc->mutex);
+
+    return buffer;
 }
 
 /* Causes the next call to poll_block() to wake up when a packet may be ready
  * to be received by vconn_recv() on 'rc'.  */
 void
 rconn_recv_wait(struct rconn *rc)
+    OVS_EXCLUDED(rc->mutex)
 {
+    ovs_mutex_lock(&rc->mutex);
     if (rc->vconn) {
         vconn_wait(rc->vconn, WAIT_RECV);
     }
+    ovs_mutex_unlock(&rc->mutex);
 }
 
-/* Sends 'b' on 'rc'.  Returns 0 if successful, or ENOTCONN if 'rc' is not
- * currently connected.  Takes ownership of 'b'.
- *
- * If 'counter' is non-null, then 'counter' will be incremented while the
- * packet is in flight, then decremented when it has been sent (or discarded
- * due to disconnection).  Because 'b' may be sent (or discarded) before this
- * function returns, the caller may not be able to observe any change in
- * 'counter'.
- *
- * There is no rconn_send_wait() function: an rconn has a send queue that it
- * takes care of sending if you call rconn_run(), which will have the side
- * effect of waking up poll_block(). */
-int
-rconn_send(struct rconn *rc, struct ofpbuf *b,
+static int
+rconn_send__(struct rconn *rc, struct ofpbuf *b,
            struct rconn_packet_counter *counter)
+    OVS_REQUIRES(rc->mutex)
 {
     if (rconn_is_connected(rc)) {
         COVERAGE_INC(rconn_queued);
@@ -632,6 +742,32 @@ rconn_send(struct rconn *rc, struct ofpbuf *b,
     }
 }
 
+/* Sends 'b' on 'rc'.  Returns 0 if successful, or ENOTCONN if 'rc' is not
+ * currently connected.  Takes ownership of 'b'.
+ *
+ * If 'counter' is non-null, then 'counter' will be incremented while the
+ * packet is in flight, then decremented when it has been sent (or discarded
+ * due to disconnection).  Because 'b' may be sent (or discarded) before this
+ * function returns, the caller may not be able to observe any change in
+ * 'counter'.
+ *
+ * There is no rconn_send_wait() function: an rconn has a send queue that it
+ * takes care of sending if you call rconn_run(), which will have the side
+ * effect of waking up poll_block(). */
+int
+rconn_send(struct rconn *rc, struct ofpbuf *b,
+           struct rconn_packet_counter *counter)
+    OVS_EXCLUDED(rc->mutex)
+{
+    int error;
+
+    ovs_mutex_lock(&rc->mutex);
+    error = rconn_send__(rc, b, counter);
+    ovs_mutex_unlock(&rc->mutex);
+
+    return error;
+}
+
 /* Sends 'b' on 'rc'.  Increments 'counter' while the packet is in flight; it
  * will be decremented when it has been sent (or discarded due to
  * disconnection).  Returns 0 if successful, EAGAIN if 'counter->n' is already
@@ -647,14 +783,21 @@ rconn_send(struct rconn *rc, struct ofpbuf *b,
 int
 rconn_send_with_limit(struct rconn *rc, struct ofpbuf *b,
                       struct rconn_packet_counter *counter, int queue_limit)
+    OVS_EXCLUDED(rc->mutex)
 {
-    if (counter->n_packets < queue_limit) {
-        return rconn_send(rc, b, counter);
+    int error;
+
+    ovs_mutex_lock(&rc->mutex);
+    if (rconn_packet_counter_n_packets(counter) < queue_limit) {
+        error = rconn_send__(rc, b, counter);
     } else {
         COVERAGE_INC(rconn_overflow);
         ofpbuf_delete(b);
-        return EAGAIN;
+        error = EAGAIN;
     }
+    ovs_mutex_unlock(&rc->mutex);
+
+    return error;
 }
 
 /* Returns the total number of packets successfully sent on the underlying
@@ -670,7 +813,9 @@ rconn_packets_sent(const struct rconn *rc)
  * and received on 'rconn' will be copied.  'rc' takes ownership of 'vconn'. */
 void
 rconn_add_monitor(struct rconn *rc, struct vconn *vconn)
+    OVS_EXCLUDED(rc->mutex)
 {
+    ovs_mutex_lock(&rc->mutex);
     if (rc->n_monitors < ARRAY_SIZE(rc->monitors)) {
         VLOG_INFO("new monitor connection from %s", vconn_get_name(vconn));
         rc->monitors[rc->n_monitors++] = vconn;
@@ -679,6 +824,7 @@ rconn_add_monitor(struct rconn *rc, struct vconn *vconn)
                  vconn_get_name(vconn));
         vconn_close(vconn);
     }
+    ovs_mutex_unlock(&rc->mutex);
 }
 
 /* Returns 'rc''s name.  This is a name for human consumption, appropriate for
@@ -693,9 +839,12 @@ rconn_get_name(const struct rconn *rc)
 /* Sets 'rc''s name to 'new_name'. */
 void
 rconn_set_name(struct rconn *rc, const char *new_name)
+    OVS_EXCLUDED(rc->mutex)
 {
+    ovs_mutex_lock(&rc->mutex);
     free(rc->name);
     rc->name = xstrdup(new_name);
+    ovs_mutex_unlock(&rc->mutex);
 }
 
 /* Returns 'rc''s target.  This is intended to be a string that may be passed
@@ -721,13 +870,27 @@ rconn_is_connected(const struct rconn *rconn)
     return is_connected_state(rconn->state);
 }
 
+static bool
+rconn_is_admitted__(const struct rconn *rconn)
+    OVS_REQUIRES(rconn->mutex)
+{
+    return (rconn_is_connected(rconn)
+            && rconn->last_admitted >= rconn->last_connected);
+}
+
 /* Returns true if 'rconn' is connected and thought to have been accepted by
  * the peer's admission-control policy. */
 bool
 rconn_is_admitted(const struct rconn *rconn)
+    OVS_EXCLUDED(rconn->mutex)
 {
-    return (rconn_is_connected(rconn)
-            && rconn->last_admitted >= rconn->last_connected);
+    bool admitted;
+
+    ovs_mutex_lock(&rconn->mutex);
+    admitted = rconn_is_admitted__(rconn);
+    ovs_mutex_unlock(&rconn->mutex);
+
+    return admitted;
 }
 
 /* Returns 0 if 'rconn' is currently connected and considered to have been
@@ -735,8 +898,17 @@ rconn_is_admitted(const struct rconn *rconn)
  * seconds since 'rconn' was last in such a state. */
 int
 rconn_failure_duration(const struct rconn *rconn)
+    OVS_EXCLUDED(rconn->mutex)
 {
-    return rconn_is_admitted(rconn) ? 0 : time_now() - rconn->last_admitted;
+    int duration;
+
+    ovs_mutex_lock(&rconn->mutex);
+    duration = (rconn_is_admitted__(rconn)
+                ? 0
+                : time_now() - rconn->last_admitted);
+    ovs_mutex_unlock(&rconn->mutex);
+
+    return duration;
 }
 
 /* Returns the IP address of the peer, or 0 if the peer's IP address is not
@@ -768,16 +940,37 @@ rconn_get_local_ip(const struct rconn *rconn)
  * connection does not contain a port or if the port is not known. */
 ovs_be16
 rconn_get_local_port(const struct rconn *rconn)
+    OVS_EXCLUDED(rconn->mutex)
 {
-    return rconn->vconn ? vconn_get_local_port(rconn->vconn) : 0;
+    ovs_be16 port;
+
+    ovs_mutex_lock(&rconn->mutex);
+    port = rconn->vconn ? vconn_get_local_port(rconn->vconn) : 0;
+    ovs_mutex_unlock(&rconn->mutex);
+
+    return port;
+}
+
+static int
+rconn_get_version__(const struct rconn *rconn)
+    OVS_REQUIRES(rconn->mutex)
+{
+    return rconn->vconn ? vconn_get_version(rconn->vconn) : -1;
 }
 
 /* Returns the OpenFlow version negotiated with the peer, or -1 if there is
  * currently no connection or if version negotiation is not yet complete. */
 int
 rconn_get_version(const struct rconn *rconn)
+    OVS_EXCLUDED(rconn->mutex)
 {
-    return rconn->vconn ? vconn_get_version(rconn->vconn) : -1;
+    int version;
+
+    ovs_mutex_lock(&rconn->mutex);
+    version = rconn_get_version__(rconn);
+    ovs_mutex_unlock(&rconn->mutex);
+
+    return version;
 }
 
 /* Returns the total number of packets successfully received by the underlying
@@ -839,15 +1032,25 @@ rconn_get_last_error(const struct rconn *rc)
 /* Returns the number of messages queued for transmission on 'rc'. */
 unsigned int
 rconn_count_txqlen(const struct rconn *rc)
+    OVS_EXCLUDED(rc->mutex)
 {
-    return list_size(&rc->txq);
+    unsigned int len;
+
+    ovs_mutex_lock(&rc->mutex);
+    len = list_size(&rc->txq);
+    ovs_mutex_unlock(&rc->mutex);
+
+    return len;
 }
 
 struct rconn_packet_counter *
 rconn_packet_counter_create(void)
 {
     struct rconn_packet_counter *c = xzalloc(sizeof *c);
+    ovs_mutex_init(&c->mutex);
+    ovs_mutex_lock(&c->mutex);
     c->ref_cnt = 1;
+    ovs_mutex_unlock(&c->mutex);
     return c;
 }
 
@@ -855,8 +1058,15 @@ void
 rconn_packet_counter_destroy(struct rconn_packet_counter *c)
 {
     if (c) {
+        bool dead;
+
+        ovs_mutex_lock(&c->mutex);
         ovs_assert(c->ref_cnt > 0);
-        if (!--c->ref_cnt && !c->n_packets) {
+        dead = !--c->ref_cnt && !c->n_packets;
+        ovs_mutex_unlock(&c->mutex);
+
+        if (dead) {
+            ovs_mutex_destroy(&c->mutex);
             free(c);
         }
     }
@@ -865,25 +1075,56 @@ rconn_packet_counter_destroy(struct rconn_packet_counter *c)
 void
 rconn_packet_counter_inc(struct rconn_packet_counter *c, unsigned int n_bytes)
 {
+    ovs_mutex_lock(&c->mutex);
     c->n_packets++;
     c->n_bytes += n_bytes;
+    ovs_mutex_unlock(&c->mutex);
 }
 
 void
 rconn_packet_counter_dec(struct rconn_packet_counter *c, unsigned int n_bytes)
 {
-    ovs_assert(c->n_packets > 0);
-    ovs_assert(c->n_bytes >= n_bytes);
+    bool dead = false;
 
-    c->n_bytes -= n_bytes;
+    ovs_mutex_lock(&c->mutex);
+    ovs_assert(c->n_packets > 0);
+    ovs_assert(c->n_packets == 1
+               ? c->n_bytes == n_bytes
+               : c->n_bytes > n_bytes);
     c->n_packets--;
-    if (!c->n_packets) {
-        ovs_assert(!c->n_bytes);
-        if (!c->ref_cnt) {
-            free(c);
-        }
+    c->n_bytes -= n_bytes;
+    dead = !c->n_packets && !c->ref_cnt;
+    ovs_mutex_unlock(&c->mutex);
+
+    if (dead) {
+        ovs_mutex_destroy(&c->mutex);
+        free(c);
     }
 }
+
+unsigned int
+rconn_packet_counter_n_packets(const struct rconn_packet_counter *c)
+{
+    unsigned int n;
+
+    ovs_mutex_lock(&c->mutex);
+    n = c->n_packets;
+    ovs_mutex_unlock(&c->mutex);
+
+    return n;
+}
+
+unsigned int
+rconn_packet_counter_n_bytes(const struct rconn_packet_counter *c)
+{
+    unsigned int n;
+
+    ovs_mutex_lock(&c->mutex);
+    n = c->n_bytes;
+    ovs_mutex_unlock(&c->mutex);
+
+    return n;
+}
 
 /* Set rc->target and rc->name to 'target' and 'name', respectively.  If 'name'
  * is null, 'target' is used.
@@ -892,6 +1133,7 @@ rconn_packet_counter_dec(struct rconn_packet_counter *c, unsigned int n_bytes)
  * the target also likely changes these values. */
 static void
 rconn_set_target__(struct rconn *rc, const char *target, const char *name)
+    OVS_REQUIRES(rc->mutex)
 {
     free(rc->name);
     rc->name = xstrdup(name ? name : target);
@@ -906,6 +1148,7 @@ rconn_set_target__(struct rconn *rc, const char *target, const char *name)
  * otherwise a positive errno value. */
 static int
 try_send(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     struct ofpbuf *msg = ofpbuf_from_list(rc->txq.next);
     unsigned int n_bytes = msg->size;
@@ -939,6 +1182,7 @@ try_send(struct rconn *rc)
  * normally. */
 static void
 report_error(struct rconn *rc, int error)
+    OVS_REQUIRES(rc->mutex)
 {
     if (error == EOF) {
         /* If 'rc' isn't reliable, then we don't really expect this connection
@@ -947,7 +1191,8 @@ report_error(struct rconn *rc, int error)
         enum vlog_level level = rc->reliable ? VLL_INFO : VLL_DBG;
         VLOG(level, "%s: connection closed by peer", rc->name);
     } else {
-        VLOG_WARN("%s: connection dropped (%s)", rc->name, strerror(error));
+        VLOG_WARN("%s: connection dropped (%s)",
+                  rc->name, ovs_strerror(error));
     }
 }
 
@@ -963,6 +1208,7 @@ report_error(struct rconn *rc, int error)
  */
 static void
 disconnect(struct rconn *rc, int error)
+    OVS_REQUIRES(rc->mutex)
 {
     rc->last_error = error;
     if (rc->reliable) {
@@ -993,7 +1239,7 @@ disconnect(struct rconn *rc, int error)
         state_transition(rc, S_BACKOFF);
     } else {
         rc->last_disconnected = time_now();
-        rconn_disconnect(rc);
+        rconn_disconnect__(rc);
     }
 }
 
@@ -1001,6 +1247,7 @@ disconnect(struct rconn *rc, int error)
  * counts. */
 static void
 flush_queue(struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     if (list_is_empty(&rc->txq)) {
         return;
@@ -1019,12 +1266,14 @@ flush_queue(struct rconn *rc)
 
 static unsigned int
 elapsed_in_this_state(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     return time_now() - rc->state_entered;
 }
 
 static unsigned int
 timeout(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     switch (rc->state) {
 #define STATE(NAME, VALUE) case S_##NAME: return timeout_##NAME(rc);
@@ -1037,12 +1286,14 @@ timeout(const struct rconn *rc)
 
 static bool
 timed_out(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     return time_now() >= sat_add(rc->state_entered, timeout(rc));
 }
 
 static void
 state_transition(struct rconn *rc, enum state state)
+    OVS_REQUIRES(rc->mutex)
 {
     rc->seqno += (rc->state == S_ACTIVE) != (state == S_ACTIVE);
     if (is_connected_state(state) && !is_connected_state(rc->state)) {
@@ -1056,8 +1307,19 @@ state_transition(struct rconn *rc, enum state state)
     rc->state_entered = time_now();
 }
 
+static void
+close_monitor(struct rconn *rc, size_t idx, int retval)
+    OVS_REQUIRES(rc->mutex)
+{
+    VLOG_DBG("%s: closing monitor connection to %s: %s",
+             rconn_get_name(rc), vconn_get_name(rc->monitors[idx]),
+             ovs_retval_to_string(retval));
+    rc->monitors[idx] = rc->monitors[--rc->n_monitors];
+}
+
 static void
 copy_to_monitor(struct rconn *rc, const struct ofpbuf *b)
+    OVS_REQUIRES(rc->mutex)
 {
     struct ofpbuf *clone = NULL;
     int retval;
@@ -1073,10 +1335,7 @@ copy_to_monitor(struct rconn *rc, const struct ofpbuf *b)
         if (!retval) {
             clone = NULL;
         } else if (retval != EAGAIN) {
-            VLOG_DBG("%s: closing monitor connection to %s: %s",
-                     rconn_get_name(rc), vconn_get_name(vconn),
-                     strerror(retval));
-            rc->monitors[i] = rc->monitors[--rc->n_monitors];
+            close_monitor(rc, i, retval);
             continue;
         }
         i++;
@@ -1116,21 +1375,14 @@ is_admitted_msg(const struct ofpbuf *b)
     case OFPTYPE_QUEUE_GET_CONFIG_REPLY:
     case OFPTYPE_GET_ASYNC_REQUEST:
     case OFPTYPE_GET_ASYNC_REPLY:
-    case OFPTYPE_METER_MOD:
-    case OFPTYPE_GROUP_REQUEST:
-    case OFPTYPE_GROUP_REPLY:
-    case OFPTYPE_GROUP_DESC_REQUEST:
-    case OFPTYPE_GROUP_DESC_REPLY:
-    case OFPTYPE_GROUP_FEATURES_REQUEST:
-    case OFPTYPE_GROUP_FEATURES_REPLY:
-    case OFPTYPE_METER_REQUEST:
-    case OFPTYPE_METER_REPLY:
-    case OFPTYPE_METER_CONFIG_REQUEST:
-    case OFPTYPE_METER_CONFIG_REPLY:
-    case OFPTYPE_METER_FEATURES_REQUEST:
-    case OFPTYPE_METER_FEATURES_REPLY:
-    case OFPTYPE_TABLE_FEATURES_REQUEST:
-    case OFPTYPE_TABLE_FEATURES_REPLY:
+    case OFPTYPE_GROUP_STATS_REQUEST:
+    case OFPTYPE_GROUP_STATS_REPLY:
+    case OFPTYPE_GROUP_DESC_STATS_REQUEST:
+    case OFPTYPE_GROUP_DESC_STATS_REPLY:
+    case OFPTYPE_GROUP_FEATURES_STATS_REQUEST:
+    case OFPTYPE_GROUP_FEATURES_STATS_REPLY:
+    case OFPTYPE_TABLE_FEATURES_STATS_REQUEST:
+    case OFPTYPE_TABLE_FEATURES_STATS_REPLY:
         return false;
 
     case OFPTYPE_PACKET_IN:
@@ -1139,6 +1391,7 @@ is_admitted_msg(const struct ofpbuf *b)
     case OFPTYPE_PACKET_OUT:
     case OFPTYPE_FLOW_MOD:
     case OFPTYPE_PORT_MOD:
+    case OFPTYPE_METER_MOD:
     case OFPTYPE_BARRIER_REQUEST:
     case OFPTYPE_BARRIER_REPLY:
     case OFPTYPE_DESC_STATS_REQUEST:
@@ -1155,6 +1408,12 @@ is_admitted_msg(const struct ofpbuf *b)
     case OFPTYPE_QUEUE_STATS_REPLY:
     case OFPTYPE_PORT_DESC_STATS_REQUEST:
     case OFPTYPE_PORT_DESC_STATS_REPLY:
+    case OFPTYPE_METER_STATS_REQUEST:
+    case OFPTYPE_METER_STATS_REPLY:
+    case OFPTYPE_METER_CONFIG_STATS_REQUEST:
+    case OFPTYPE_METER_CONFIG_STATS_REPLY:
+    case OFPTYPE_METER_FEATURES_STATS_REQUEST:
+    case OFPTYPE_METER_FEATURES_STATS_REPLY:
     case OFPTYPE_ROLE_REQUEST:
     case OFPTYPE_ROLE_REPLY:
     case OFPTYPE_SET_FLOW_FORMAT:
@@ -1178,6 +1437,7 @@ is_admitted_msg(const struct ofpbuf *b)
  * successuflly connected in too long. */
 static bool
 rconn_logging_connection_attempts__(const struct rconn *rc)
+    OVS_REQUIRES(rc->mutex)
 {
     return rc->backoff < rc->max_backoff;
 }
