@@ -57,10 +57,14 @@ BUILD_ASSERT_DECL(TYPE_IS_SIGNED(time_t));
 #endif
 #endif /* ifndef CACHE_TIME */
 
-void time_disable_restart(void);
-void time_enable_restart(void);
+struct tm_msec {
+  struct tm tm;
+  int msec;
+};
+
 void time_postfork(void);
 void time_refresh(void);
+
 time_t time_now(void);
 time_t time_wall(void);
 long long int time_msec(void);
@@ -70,15 +74,18 @@ void time_wall_timespec(struct timespec *);
 void time_alarm(unsigned int secs);
 int time_poll(struct pollfd *, int n_pollfds, long long int timeout_when,
               int *elapsed);
-bool time_cached(void);
 
 long long int timespec_to_msec(const struct timespec *);
 long long int timeval_to_msec(const struct timeval *);
 
+struct tm_msec *localtime_msec(long long int now, struct tm_msec *result);
+struct tm_msec *gmtime_msec(long long int now, struct tm_msec *result);
+size_t strftime_msec(char *s, size_t max, const char *format,
+                     const struct tm_msec *);
 void xgettimeofday(struct timeval *);
+void xclock_gettime(clock_t, struct timespec *);
 
 int get_cpu_usage(void);
-void format_backtraces(struct ds *, size_t min_count);
 
 long long int time_boot_msec(void);
 
