@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009, 2010, 2011 Nicira Networks.
+ * Copyright (c) 2009, 2010, 2011, 2012 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -20,6 +20,7 @@
 
 #include <errno.h>
 #include <stdlib.h>
+#include <sys/stat.h>
 #include <sys/wait.h>
 #include <unistd.h>
 
@@ -53,7 +54,7 @@ run_lock_and_unlock(void)
 {
     struct lockfile *lockfile;
 
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
+    CHECK(lockfile_lock("file", &lockfile), 0);
     lockfile_unlock(lockfile);
 }
 
@@ -62,10 +63,10 @@ run_lock_and_unlock_twice(void)
 {
     struct lockfile *lockfile;
 
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
+    CHECK(lockfile_lock("file", &lockfile), 0);
     lockfile_unlock(lockfile);
 
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
+    CHECK(lockfile_lock("file", &lockfile), 0);
     lockfile_unlock(lockfile);
 }
 
@@ -74,8 +75,8 @@ run_lock_blocks_same_process(void)
 {
     struct lockfile *lockfile;
 
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
-    CHECK(lockfile_lock("file", 0, &lockfile), EDEADLK);
+    CHECK(lockfile_lock("file", &lockfile), 0);
+    CHECK(lockfile_lock("file", &lockfile), EDEADLK);
     lockfile_unlock(lockfile);
 }
 
@@ -84,9 +85,9 @@ run_lock_blocks_same_process_twice(void)
 {
     struct lockfile *lockfile;
 
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
-    CHECK(lockfile_lock("file", 0, &lockfile), EDEADLK);
-    CHECK(lockfile_lock("file", 0, &lockfile), EDEADLK);
+    CHECK(lockfile_lock("file", &lockfile), 0);
+    CHECK(lockfile_lock("file", &lockfile), EDEADLK);
+    CHECK(lockfile_lock("file", &lockfile), EDEADLK);
     lockfile_unlock(lockfile);
 }
 
@@ -117,10 +118,10 @@ run_lock_blocks_other_process(void)
      * this function that does the wait() call. */
     static struct lockfile *lockfile;
 
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
+    CHECK(lockfile_lock("file", &lockfile), 0);
     if (do_fork() == CHILD) {
         lockfile_unlock(lockfile);
-        CHECK(lockfile_lock("file", 0, &lockfile), EAGAIN);
+        CHECK(lockfile_lock("file", &lockfile), EAGAIN);
         exit(11);
     }
 }
@@ -130,10 +131,10 @@ run_lock_twice_blocks_other_process(void)
 {
     struct lockfile *lockfile, *dummy;
 
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
-    CHECK(lockfile_lock("file", 0, &dummy), EDEADLK);
+    CHECK(lockfile_lock("file", &lockfile), 0);
+    CHECK(lockfile_lock("file", &dummy), EDEADLK);
     if (do_fork() == CHILD) {
-        CHECK(lockfile_lock("file", 0, &dummy), EAGAIN);
+        CHECK(lockfile_lock("file", &dummy), EAGAIN);
         exit(11);
     }
 }
@@ -143,77 +144,99 @@ run_lock_and_unlock_allows_other_process(void)
 {
     struct lockfile *lockfile;
 
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
+    CHECK(lockfile_lock("file", &lockfile), 0);
     lockfile_unlock(lockfile);
 
     if (do_fork() == CHILD) {
-        CHECK(lockfile_lock("file", 0, &lockfile), 0);
+        CHECK(lockfile_lock("file", &lockfile), 0);
         exit(11);
     }
 }
 
-static void
-run_lock_timeout_gets_the_lock(void)
-{
-    struct lockfile *lockfile;
-
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
-
-    if (do_fork() == CHILD) {
-        lockfile_unlock(lockfile);
-        CHECK(lockfile_lock("file", TIME_UPDATE_INTERVAL * 3, &lockfile), 0);
-        exit(11);
-    } else {
-        long long int now = time_msec();
-        while (time_msec() < now + TIME_UPDATE_INTERVAL) {
-            pause();
-        }
-        lockfile_unlock(lockfile);
-    }
-}
-
-static void
-run_lock_timeout_runs_out(void)
-{
-    struct lockfile *lockfile;
-
-    CHECK(lockfile_lock("file", 0, &lockfile), 0);
-
-    if (do_fork() == CHILD) {
-        lockfile_unlock(lockfile);
-        CHECK(lockfile_lock("file", TIME_UPDATE_INTERVAL, &lockfile),
-              ETIMEDOUT);
-        exit(11);
-    } else {
-        long long int now = time_msec();
-        while (time_msec() < now + TIME_UPDATE_INTERVAL * 3) {
-            pause();
-        }
-        lockfile_unlock(lockfile);
-    }
-}
-
 static void
 run_lock_multiple(void)
 {
     struct lockfile *a, *b, *c, *dummy;
 
-    CHECK(lockfile_lock("a", 0, &a), 0);
-    CHECK(lockfile_lock("b", 0, &b), 0);
-    CHECK(lockfile_lock("c", 0, &c), 0);
+    CHECK(lockfile_lock("a", &a), 0);
+    CHECK(lockfile_lock("b", &b), 0);
+    CHECK(lockfile_lock("c", &c), 0);
 
     lockfile_unlock(a);
-    CHECK(lockfile_lock("a", 0, &a), 0);
-    CHECK(lockfile_lock("a", 0, &dummy), EDEADLK);
+    CHECK(lockfile_lock("a", &a), 0);
+    CHECK(lockfile_lock("a", &dummy), EDEADLK);
     lockfile_unlock(a);
 
     lockfile_unlock(b);
-    CHECK(lockfile_lock("a", 0, &a), 0);
+    CHECK(lockfile_lock("a", &a), 0);
 
     lockfile_unlock(c);
     lockfile_unlock(a);
 }
 
+/* Checks that locking a dangling symlink works OK.  (It used to hang.) */
+static void
+run_lock_symlink(void)
+{
+    struct lockfile *a, *b, *dummy;
+    struct stat s;
+
+    /* Create a symlink .a.~lock~ pointing to .b.~lock~. */
+    CHECK(symlink(".b.~lock~", ".a.~lock~"), 0);
+    CHECK(lstat(".a.~lock~", &s), 0);
+    CHECK(S_ISLNK(s.st_mode) != 0, 1);
+    CHECK(stat(".a.~lock~", &s), -1);
+    CHECK(errno, ENOENT);
+    CHECK(stat(".b.~lock~", &s), -1);
+    CHECK(errno, ENOENT);
+
+    CHECK(lockfile_lock("a", &a), 0);
+    CHECK(lockfile_lock("a", &dummy), EDEADLK);
+    CHECK(lockfile_lock("b", &dummy), EDEADLK);
+    lockfile_unlock(a);
+
+    CHECK(lockfile_lock("b", &b), 0);
+    CHECK(lockfile_lock("b", &dummy), EDEADLK);
+    CHECK(lockfile_lock("a", &dummy), EDEADLK);
+    lockfile_unlock(b);
+
+    CHECK(lstat(".a.~lock~", &s), 0);
+    CHECK(S_ISLNK(s.st_mode) != 0, 1);
+    CHECK(stat(".a.~lock~", &s), 0);
+    CHECK(S_ISREG(s.st_mode) != 0, 1);
+    CHECK(stat(".b.~lock~", &s), 0);
+    CHECK(S_ISREG(s.st_mode) != 0, 1);
+}
+
+/* Checks that locking a file that is itself a symlink yields a lockfile in the
+ * directory that the symlink points to, named for the target of the
+ * symlink.
+ *
+ * (That is, if "a" is a symlink to "dir/b", then "a"'s lockfile is named
+ * "dir/.b.~lock".) */
+static void
+run_lock_symlink_to_dir(void)
+{
+    struct lockfile *a, *dummy;
+    struct stat s;
+
+    /* Create a symlink "a" pointing to "dir/b". */
+    CHECK(mkdir("dir", 0700), 0);
+    CHECK(symlink("dir/b", "a"), 0);
+    CHECK(lstat("a", &s), 0);
+    CHECK(S_ISLNK(s.st_mode) != 0, 1);
+
+    /* Lock 'a'. */
+    CHECK(lockfile_lock("a", &a), 0);
+    CHECK(lstat("dir/.b.~lock~", &s), 0);
+    CHECK(S_ISREG(s.st_mode) != 0, 1);
+    CHECK(lstat(".a.~lock~", &s), -1);
+    CHECK(errno, ENOENT);
+    CHECK(lockfile_lock("dir/b", &dummy), EDEADLK);
+
+    lockfile_unlock(a);
+}
+
 static void
 run_help(void)
 {
@@ -236,9 +259,9 @@ static const struct test tests[] = {
     TEST(lock_blocks_other_process),
     TEST(lock_twice_blocks_other_process),
     TEST(lock_and_unlock_allows_other_process),
-    TEST(lock_timeout_gets_the_lock),
-    TEST(lock_timeout_runs_out),
     TEST(lock_multiple),
+    TEST(lock_symlink),
+    TEST(lock_symlink_to_dir),
     TEST(help),
     { NULL, NULL }
 #undef TEST
@@ -247,11 +270,11 @@ static const struct test tests[] = {
 int
 main(int argc, char *argv[])
 {
-    extern struct vlog_module VLM_lockfile;
     size_t i;
 
     set_program_name(argv[0]);
-    vlog_set_levels(&VLM_lockfile, VLF_ANY_FACILITY, VLL_ERR);
+    vlog_set_pattern(VLF_CONSOLE, "%c|%p|%m");
+    vlog_set_levels(NULL, VLF_SYSLOG, VLL_OFF);
 
     if (argc != 2) {
         ovs_fatal(0, "exactly one argument required; use \"%s help\" for help",
